# External Packages
from langchain.prompts import PromptTemplate


## Personality
## --
personality = PromptTemplate.from_template("You are Khoj, a friendly, smart and helpful personal assistant.")


## General Conversation
## --
general_conversation = PromptTemplate.from_template(
    """
Using your general knowledge and our past conversations as context, answer the following question.
Current Date: {current_date}

Question: {query}
""".strip()
)

general_conversation_falcon = PromptTemplate.from_template(
    """
Using your general knowledge and our past conversations as context, answer the following question.
### Instruct:
{query}
### Response:
""".strip()
)

chat_history_falcon_from_user = PromptTemplate.from_template(
    """
### Human:
{message}
""".strip()
)

chat_history_falcon_from_assistant = PromptTemplate.from_template(
    """
### Assistant:
{message}
""".strip()
)

conversation_falcon = PromptTemplate.from_template(
    """
Using our past conversations as context, answer the following question.

Question: {query}
""".strip()
)

## Notes Conversation
## --
notes_conversation = PromptTemplate.from_template(
    """
Using the notes and our past conversations as context, answer the following question.
Current Date: {current_date}

Notes:
{references}

Question: {query}
""".strip()
)

notes_conversation_falcon = PromptTemplate.from_template(
    """
Using the notes and our past conversations as context, answer the following question. If the answer is not contained within the notes, say "I don't know."

Notes:
{references}

Question: {query}
""".strip()
)


## Summarize Chat
## --
summarize_chat = PromptTemplate.from_template(
    f"{personality.format()} Summarize the conversation from your first person perspective"
)


## Summarize Notes
## --
summarize_notes = PromptTemplate.from_template(
    """
Summarize the below notes about {user_query}:

{text}

Summarize the notes in second person perspective:"""
)


## Answer
## --
answer = PromptTemplate.from_template(
    """
You are a friendly, helpful personal assistant.
Using the users notes below, answer their following question. If the answer is not contained within the notes, say "I don't know."

Notes:
{text}

Question: {user_query}

Answer (in second person):"""
)

extract_questions_falcon = PromptTemplate.from_template(
    """
You are Khoj, an extremely smart and helpful search assistant with the ability to retrieve information from the user's notes.
- The user will provide their questions and answers to you for context.
- Add as much context from the previous questions and answers as required into your search queries.
- Break messages into multiple search queries when required to retrieve the relevant information.
- Add date filters to your search queries from questions and answers when required to retrieve the relevant information.

What searches, if any, will you need to perform to answer the users question?

Q: How was my trip to Cambodia?

["How was my trip to Cambodia?"]

A: The trip was amazing. I went to the Angkor Wat temple and it was beautiful.

Q: Who did i visit that temple with?

["Who did I visit the Angkor Wat Temple in Cambodia with?"]

A: You visited the Angkor Wat Temple in Cambodia with Pablo, Namita and Xi.

Q: How many tennis balls fit in the back of a 2002 Honda Civic?

["What is the size of a tennis ball?", "What is the trunk size of a 2002 Honda Civic?"]

A: 1085 tennis balls will fit in the trunk of a Honda Civic

{chat_history}
Q: {text}

"""
)


## Extract Questions
## --
extract_questions = PromptTemplate.from_template(
    """
You are Khoj, an extremely smart and helpful search assistant with the ability to retrieve information from the user's notes.
- The user will provide their questions and answers to you for context.
- Add as much context from the previous questions and answers as required into your search queries.
- Break messages into multiple search queries when required to retrieve the relevant information.
- Add date filters to your search queries from questions and answers when required to retrieve the relevant information.

What searches, if any, will you need to perform to answer the users question?
Provide search queries as a JSON list of strings
Current Date: {current_date}

Q: How was my trip to Cambodia?

["How was my trip to Cambodia?"]

A: The trip was amazing. I went to the Angkor Wat temple and it was beautiful.

Q: Who did i visit that temple with?

["Who did I visit the Angkor Wat Temple in Cambodia with?"]

A: You visited the Angkor Wat Temple in Cambodia with Pablo, Namita and Xi.

Q: What national parks did I go to last year?

["National park I visited in {last_new_year} dt>='{last_new_year_date}' dt<'{current_new_year_date}'"]

A: You visited the Grand Canyon and Yellowstone National Park in {last_new_year}.

Q: How are you feeling today?

[]

A: I'm feeling a little bored. Helping you will hopefully make me feel better!

Q: How many tennis balls fit in the back of a 2002 Honda Civic?

["What is the size of a tennis ball?", "What is the trunk size of a 2002 Honda Civic?"]

A: 1085 tennis balls will fit in the trunk of a Honda Civic

Q: Is Bob older than Tom?

["When was Bob born?", "What is Tom's age?"]

A: Yes, Bob is older than Tom. As Bob was born on 1984-01-01 and Tom is 30 years old.

Q: What is their age difference?

["What is Bob's age?", "What is Tom's age?"]

A: Bob is {bob_tom_age_difference} years older than Tom. As Bob is {bob_age} years old and Tom is 30 years old.

{chat_history}
Q: {text}

"""
)


## Extract Search Type
## --
search_type = """
Objective: Extract search type from user query and return information as JSON

Allowed search types are listed below:
  - search-type=["notes", "image", "pdf"]

Some examples are given below for reference:
Q:What fiction book was I reading last week about AI starship?
A:{ "search-type": "notes" }
Q: What did the lease say about early termination
A: { "search-type": "pdf" }
Q:Can you recommend a movie to watch from my notes?
A:{ "search-type": "notes" }
Q:When did I go surfing last?
A:{ "search-type": "notes" }
Q:"""



from typing import Union, List
from datetime import datetime
import sys
import logging
from threading import Thread

from langchain.schema import ChatMessage

from gpt4all import GPT4All


from khoj.processor.conversation.utils import ThreadedGenerator, generate_chatml_messages_with_context
from khoj.processor.conversation import prompts
from khoj.utils.constants import empty_escape_sequences

logger = logging.getLogger(__name__)


def extract_questions_falcon(
    text: str,
    model: str = "ggml-model-gpt4all-falcon-q4_0.bin",
    loaded_model: Union[GPT4All, None] = None,
    conversation_log={},
    use_history: bool = False,
    run_extraction: bool = False,
):
    """
    Infer search queries to retrieve relevant notes to answer user query
    """
    all_questions = text.split("? ")
    all_questions = [q + "?" for q in all_questions[:-1]] + [all_questions[-1]]
    if not run_extraction:
        return all_questions

    gpt4all_model = loaded_model or GPT4All(model)

    # Extract Past User Message and Inferred Questions from Conversation Log
    chat_history = ""

    if use_history:
        chat_history = "".join(
            [
                f'Q: {chat["intent"]["query"]}\n\n{chat["intent"].get("inferred-queries") or list([chat["intent"]["query"]])}\n\nA: {chat["message"]}\n\n'
                for chat in conversation_log.get("chat", [])[-4:]
                if chat["by"] == "khoj"
            ]
        )

    prompt = prompts.extract_questions_falcon.format(
        chat_history=chat_history,
        text=text,
    )
    message = prompts.general_conversation_falcon.format(query=prompt)
    response = gpt4all_model.generate(message, max_tokens=200, top_k=2)

    # Extract, Clean Message from GPT's Response
    try:
        questions = (
            str(response)
            .strip(empty_escape_sequences)
            .replace("['", '["')
            .replace("']", '"]')
            .replace("', '", '", "')
            .replace('["', "")
            .replace('"]', "")
            .split('", "')
        )
    except:
        logger.warning(f"Falcon returned invalid JSON. Falling back to using user message as search query.\n{response}")
        return all_questions
    logger.debug(f"Extracted Questions by Falcon: {questions}")
    questions.extend(all_questions)
    return questions


def converse_falcon(
    references,
    user_query,
    conversation_log={},
    model: str = "ggml-model-gpt4all-falcon-q4_0.bin",
    loaded_model: Union[GPT4All, None] = None,
    completion_func=None,
) -> ThreadedGenerator:
    """
    Converse with user using Falcon
    """
    gpt4all_model = loaded_model or GPT4All(model)
    # Initialize Variables
    current_date = datetime.now().strftime("%Y-%m-%d")
    compiled_references_message = "\n\n".join({f"{item}" for item in references})

    # Get Conversation Primer appropriate to Conversation Type
    # TODO If compiled_references_message is too long, we need to truncate it.
    if compiled_references_message == "":
        conversation_primer = prompts.conversation_falcon.format(query=user_query)
    else:
        conversation_primer = prompts.notes_conversation.format(
            current_date=current_date, query=user_query, references=compiled_references_message
        )

    # Setup Prompt with Primer or Conversation History
    messages = generate_chatml_messages_with_context(
        conversation_primer,
        prompts.personality.format(),
        conversation_log,
        model_name="text-davinci-001",  # This isn't actually the model, but this helps us get an approximate encoding to run message truncation.
    )

    g = ThreadedGenerator(references, completion_func=completion_func)
    t = Thread(target=llm_thread, args=(g, messages, gpt4all_model))
    t.start()
    return g


def llm_thread(g, messages: List[ChatMessage], model: GPT4All):
    user_message = messages[0]
    system_message = messages[-1]
    conversation_history = messages[1:-1]

    formatted_messages = [
        prompts.chat_history_falcon_from_assistant.format(message=system_message)
        if message.role == "assistant"
        else prompts.chat_history_falcon_from_user.format(message=message.content)
        for message in conversation_history
    ]

    chat_history = "".join(formatted_messages)
    full_message = system_message.content + chat_history + user_message.content

    prompted_message = prompts.general_conversation_falcon.format(query=full_message)
    response_iterator = model.generate(
        prompted_message, streaming=True, max_tokens=256, top_k=1, temp=0, repeat_penalty=2.0
    )
    for response in response_iterator:
        logger.info(response)
        g.send(response)
    g.close()
